/***************
 *  TI92TC.C
 ***************
Copyright (c) 1998 David Kuehling

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

 * This is the sourcefile of my TI92TC - text converter. It converts TI-92
 * Graph Link Text Variables (*.92t) from and to normal DOS text files. It
 * is capable of TI-92 character set <-> ASCII convertions, and of word
 * wrapping.
 *
 * This programm was compiled and tested with the DJGPP-C C compiler,
 * which you can download from http://www.delorie.com. You will have to make
 * some modifications on the sourcecode, to compile it with other compilers.
 * This is especially neccesary due to the nested function definitions, that
 * aren't supported by most of the other C compilers.
 *
 * Last modification: september 1998
 */

/** number of current version */
#define VERSION_STR "1.01"

/****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <dos.h>
#include <errno.h>
/****************************************************************************/

FILE            *InHdl,
                *OutHdl;
unsigned char   *InFile;
char            *OutFileName;
int             WordWrap = 0,
                CharConv = 0,
                TabLen = 8;
long            InFileLength;


/**
 * Calculates a word - checksum, by adding all bytes from
 * *data to data[len-1]
 */
unsigned short GenerateCSum (unsigned char *data, long len)
{
  short csum;

  for (csum = 0; len > 0; len--)
    csum += *data++;

  return (csum);
}

/**
 * Converts the *.92t - file `InHdl' to a *.txt - file `OutHdl'
 */
void Convert92T2TXT ()
{
  struct        CHAR_CONVERTION       /** TI-92 --> normal text convertion */
  {
    unsigned char *out;
    int            len;
  } charconv[] =
  {
    {"\x00", 1}, {"\x01", 1}, {"\x02", 1}, {"\x03", 1},
    {"\x04", 1}, {"\x05", 1}, {"\x06", 1}, {"\x07", 1},
    {"\x08", 1}, {"\x09", 1}, {"\x0A", 1}, {"\x0B", 1},
    {"\x0C", 1}, {"\x0D", 1}, {"\x0E", 1}, {"\xFB", 1},

    {"\xFE", 1}, {"\x11", 1}, {"\x10", 1}, {"\x1E", 1},
    {"\x1F", 1}, {"\x1B", 1}, {  "->", 1}, {  "<-", 1},
    {"\x19", 1}, {"\x11", 1}, {"\x10", 1}, {"\x18", 1},
    {   "U", 1}, {"\xEF", 1}, {   "C", 1}, {"\xEE", 1},

    {"\x20", 1}, {"\x21", 1}, {"\x22", 1}, {"\x23", 1},
    {"\x24", 1}, {"\x25", 1}, {"\x26", 1}, {"\x27", 1},
    {"\x28", 1}, {"\x29", 1}, {"\x2A", 1}, {"\x2B", 1},
    {"\x2C", 1}, {"\x2D", 1}, {"\x2E", 1}, {"\x2F", 1},

    {"\x30", 1}, {"\x31", 1}, {"\x32", 1}, {"\x33", 1},
    {"\x34", 1}, {"\x35", 1}, {"\x36", 1}, {"\x37", 1},
    {"\x38", 1}, {"\x39", 1}, {"\x3A", 1}, {"\x3B", 1},
    {"\x3C", 1}, {"\x3D", 1}, {"\x3E", 1}, {"\x3F", 1},

    {"\x40", 1}, {"\x41", 1}, {"\x42", 1}, {"\x43", 1},
    {"\x44", 1}, {"\x45", 1}, {"\x46", 1}, {"\x47", 1},
    {"\x48", 1}, {"\x49", 1}, {"\x4A", 1}, {"\x4B", 1},
    {"\x4C", 1}, {"\x4D", 1}, {"\x4E", 1}, {"\x4F", 1},

    {"\x50", 1}, {"\x51", 1}, {"\x52", 1}, {"\x53", 1},
    {"\x54", 1}, {"\x55", 1}, {"\x56", 1}, {"\x57", 1},
    {"\x58", 1}, {"\x59", 1}, {"\x5A", 1}, {"\x5B", 1},
    {"\x5C", 1}, {"\x5D", 1}, {"\x5E", 1}, {"\x5F", 1},

    {"\x60", 1}, {"\x61", 1}, {"\x62", 1}, {"\x63", 1},
    {"\x64", 1}, {"\x65", 1}, {"\x66", 1}, {"\x67", 1},
    {"\x68", 1}, {"\x69", 1}, {"\x6A", 1}, {"\x6B", 1},
    {"\x6C", 1}, {"\x6D", 1}, {"\x6E", 1}, {"\x6F", 1},

    {"\x70", 1}, {"\x71", 1}, {"\x72", 1}, {"\x73", 1},
    {"\x74", 1}, {"\x75", 1}, {"\x76", 1}, {"\x77", 1},
    {"\x78", 1}, {"\x79", 1}, {"\x7A", 1}, {"\x7B", 1},
    {"\xB3", 1}, {"\x7D", 1}, {"\x7E", 1}, {"\x7F", 1},

    {"\xE0", 1}, {"\xE1", 1}, {"\xE2", 1}, {"gamma", 5},
    {"\x7F", 1}, {"\xEB", 1},{"epsilon",7},{"zeta", 4},
    {"\xE9", 1}, {"lambda",6},{  "xi", 2}, { "Pi", 2},
    {"\xE3", 1}, { "rho", 3}, {"\xE4", 1}, {"\xE5", 1},

    {"\xE7", 1}, {"\xE8", 1}, { "psi", 3}, {"\xEA", 1},
    {"omega", 5},{   "E", 1}, {   "e", 1}, {   "i", 1},
    {"\x98", 1}, {"\x99", 1}, {"\x9A", 1}, {"\x9B", 1},
    {"\xF3", 1}, { "=/=", 3}, {"\xF2", 1}, {"\x9F", 1},

    { "...", 3}, {"\xAD", 1}, {"\x9B", 1}, {"\x9C", 1},
    {"\xA4", 1}, {"\x9D", 1}, {   "|", 1}, {"\x15", 1},
    {"\xFB", 1}, { "(C)", 3}, {"\xA6", 1}, {"\xAE", 1},
    {"\xAA", 1}, {   "-", 1}, { "(R)", 3}, {   "-", 1},

    {"\xF8", 1}, {   "+", 1}, {"\xFD", 1}, {  "^3", 2},
    {"^(-1)",5}, {"\xE6", 1}, {"\x14", 1}, {"\xF9", 1},
    {   "x", 1}, {  "^1", 2}, {"\xA7", 1}, {"\xAF", 1},
    {   "d", 1},{"integrate",9},{"\xEC", 1}, {"\xA8", 1},

    {   "A", 1}, {   "A", 1}, {   "A", 1}, {   "A", 1},
    {"\x8E", 1}, {"\x8F", 1}, {"\x92", 1}, {"\x80", 1},
    {   "E", 1}, {"\x90", 1}, {   "E", 1}, {   "E", 1},
    {   "I", 1}, {   "I", 1}, {   "I", 1}, {   "I", 1},

    {   "D", 1}, {"\xA5", 1}, {   "O", 1}, {   "O", 1},
    {   "O", 1}, {   "O", 1}, {"\x99", 1}, {   "x", 1},
    {"\xED", 1}, {   "U", 1}, {   "U", 1}, {   "U", 1},
    {"\x9A", 1}, {   "Y", 1}, {"\xDE", 1}, {"\xE1", 1},

    {"\x85", 1}, {"\xA0", 1}, {"\x83", 1}, {  "a", 1},
    {"\x84", 1}, {"\x86", 1}, {"\x91", 1}, {"\x87", 1},
    {"\x8A", 1}, {"\x82", 1}, {"\x88", 1}, {"\x89", 1},
    {"\x8D", 1}, {"\xA1", 1}, {"\x8C", 1}, {"\x8B", 1},

    {"\xF0", 1}, {"\xA4", 1}, {"\x95", 1}, {"\xA2", 1},
    {"\x93", 1}, {   "o", 1}, {"\x94", 1}, {"\xF6", 1},
    {"\xED", 1}, {"\x97", 1}, {"\xA3", 1}, {"\x96", 1},
    {"\x81", 1}, {   "y", 1}, {"\xFE", 1}, {"\x98", 1},
  };
  unsigned char *txt;
  int           c,
                column;


  void InvalidWarning (char *descriptstr)
  {
    fprintf (stderr, "Warning: Invalid TI-92 Graph Link Text File (%s)\n",
                     descriptstr);
  }

  /**
   * Returns the length of a character, when it is converted (e.g.  -> ^2)
   */
  __inline__ int ChrLen (int c)
  {
    if (!CharConv) return (1);
    return (charconv[c].len);
  }

  /**
   * Writes a character/its converted sequence to OutHdl (e.g.  -> ^2)
   */
  __inline__ int PutChr (int c)
  {
    if (!CharConv) fputc (c, OutHdl);
    else fputs (charconv[c].out, OutHdl);
  }

  /**
   * Tests, whether a word wrap on the current position is necessary
   */
  int IsWordWrap ()
  {
    unsigned char *nextword,
                  *nextwordend;
    int           nwecolumn = column;

    if (!WordWrap)
      return (0);
    if (c != ' ')
    {
      if (column > WordWrap)
        return (1);
      return (0);
    }

    nextword = txt;
    while (*nextword == ' ')
      nwecolumn++, nextword++;

    if (*nextword == '\n')
    {
      txt = nextword+2;
      return (1);
    }

    nextwordend = nextword;
    while (*nextwordend != ' ' && *nextwordend != '\r' && *nextwordend != '\0')
      nwecolumn += ChrLen (*nextwordend++);

    if (nwecolumn-1 > WordWrap)
    {
      txt = nextword;
      return (1);
    }

    return (0);
  }

  int length;

  if (WordWrap == -1)  /** -1 means enabled, but default setting */
    WordWrap = 78;

  fputs ("Converting: 92T --> TXT ...", stdout);

  length = InFile[0x56] << 8 | InFile[0x57];

  if (InFileLength != *((long *) &InFile[0x4C]))
    InvalidWarning ("bad filelength specification");
  if (*((unsigned short *) &InFile[0x56+length]) != 0xE000)
    InvalidWarning ("variable type is not `TEXT'");
  if (GenerateCSum (&InFile[0x56], length+2) !=
        *((unsigned short *) &InFile[InFileLength-2]))
    InvalidWarning ("bad checksum");

  txt = &InFile[0x5B];

  for (column = 1; (c = *txt) != '\0';)
  {
    if (c == '\r')
    {
      fputc ('\r', OutHdl);
      fputc ('\n', OutHdl);
      column = 1;
      txt += 2;
    }
    else if (IsWordWrap ())
    {
      fputc ('\r', OutHdl);
      fputc ('\n', OutHdl);
      column = 1;
    }
    else
    {
      PutChr (c);
      column += ChrLen (c);
      txt++;
    }
  }

  puts (" done.");
}

/**
 * Converts the *.txt - file `InHdl' to a *.92t - file `OutHdl'
 */
void ConvertTXT292T ()
{
  void MemErr ()
  {
    fputs ("\nNot enough DPMI memory, to store output file.\n\n", stderr);
    exit (1);
  }
  void WriteErr ()
  {
    fprintf (stderr, "\nCan't write output file : %s\n\n", strerror (errno));
    exit (1);
  }

  unsigned char charconv[256] =
  {
      0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
     10,  11,  12,  13,  14,  15,  16,  17,  18,  19,
     20,  21,  22,  23,  24,  25,  26,  27,  28,  29,
     30,  31,  32,  33,  34,  35,  36,  37,  38,  39,
     40,  41,  42,  43,  44,  45,  46,  47,  48,  49,
     50,  51,  52,  53,  54,  55,  56,  57,  58,  59,
     60,  61,  62,  63,  64,  65,  66,  67,  68,  69,
     70,  71,  72,  73,  74,  75,  76,  77,  78,  79,
     80,  81,  82,  83,  84,  85,  86,  87,  88,  89,
     90,  91,  92,  93,  94,  95,  96,  97,  98,  99,
    100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
    110, 111, 112, 113, 114, 115, 116, 117, 118, 119,
    120, 121, 122, 123, 166, 125, 126, 132, 199, 252,
    233, 226, 228, 224, 229, 231, 234, 235, 232, 239,
    238, 236, 196, 197, 201, 230, 198, 244, 246, 242,
    251, 249, 255, 214, 220, 162, 163, 165,  80, 189,
    225, 237, 243, 250, 241, 209, 170, 186, 191, '?',
    172, '?', '?', 161, 171, 187, 183, '*', '#', '|',
    '+', '+', '+', '+', '+', '+', '|', '+', '+', '+',
    '+', '+', '+', '+', '+', '+', '-', '+', '+', '+',
    '+', '+', '+', '+', '+', '=', '+', '+', '+', '+',
    '+', '+', '+', '+', '+', '+', '+', '+', '+',  16,
     16,  16,  16,  16, 128, 129, 130, 140, 142, 143,
    181, 144, 145, 136, 147, 133, 190, 248,  31,  29,
    '=', '?', 158, 156, '?', '?', 247, '=', 176, 183,
    183, 168, '?', 178,  16, ' '
  };

  struct                        /* Graph Link file + TIOS var Header */
  {
    char  TI92[8];
    char  Unknown1[2];
    char  Folder[8];
    char  String[42];
    char  Unknown2[4];
    char  Name[8];
    char  Unknown3[4];
    long  GraphLinkLength;
    char  Unknown4[6];
  } header;
  struct date cdate;            /* current date (for description field) */
  struct time ctime;            /* current time (for description field) */

  char     *name,               /* pointer to name component of filename */
           ename[16],           /* fully extracted name of filename */
           *ext,                /* pointer to file extension */
           *tmp;

  unsigned char *outfile,       /* pointer to memory, where converted text is stored */
                *out,           /* pointer to current output byte */
                *in;            /* pointer to current input byte  */
  int           varlen,         /* length of TIOS-Text variable */
                column,
                c;
  short         checksum;       /* checksum of Graph Link file */

  /**
   * Tests, whether there is a word wrap on the current position
   */
  int IsWordWrap ()
  {
    unsigned char *nextword,
                  *nextwordend;
    int           nwecolumn = column;   /* next word's end column */

    if (!WordWrap)
      return (0);

    if (c != ' ' && c != '\t')
    {
      if (column > WordWrap)
        return (1);
      return (0);
    }

    nextword = in;
    while (*nextword == ' ' || *nextword == '\t')
    {
      if (*nextword++ == '\t')
        nwecolumn = (nwecolumn+7 & 0xFFF8) + 1;
      else
        nwecolumn++;

    }

    if (*nextword == '\n')
    {
      in = nextword+1;
      return (1);
    }
    if (*nextword == '\r')
    {
      in = nextword+2;
      return (1);
    }

    nextwordend = nextword;
    while (*nextwordend != ' ' && *nextwordend != '\t' &&
           *nextwordend != '\r' && *nextwordend != '\0')
      nwecolumn++, nextwordend++;

    if (nwecolumn-1 > WordWrap)
    {
      in = nextword;
      return (1);
    }

    return (0);
  }

  printf ("Converting: TXT --> 92T ...", stdout);

  if (WordWrap == -1)   /** -1 means enabled, but default setting */
    WordWrap = 37;

  if ((outfile = malloc (InFileLength*2+5)) == NULL)
    MemErr ();

  /** convert *in --> *out */
  out = outfile, 
  out += 2;                  /* TIOS variable length specifier */
  *out++ = 0;                /* Cursor position within text var */
  *out++ = 1;
  *out++ = ' ';
  for (in = InFile, column = 1, varlen = 5; c = *in;)
  {
    if (IsWordWrap ())
    {
      *out++ = '\r', *out++ = ' ';
      varlen += 2;
      column = 1;
      continue;
    }

    if (c == '\n')
    {
      *out++ = '\r', *out++ = ' ';
      varlen += 2;
      column = 1;
    }
    else if (c == '\t')
      do *out++ = ' ', varlen++; while (column++ % TabLen);
    else if (c != '\r')
    {
      *out++ = CharConv ? charconv[c] : c;
      column++;
      varlen++;
    }
    in++;
  }

  /** add TIOS filelength specification */
  outfile[0] = varlen >> 8;
  outfile[1] = varlen &0xFF;

  /** add TIOS `TEXT'-specification 0x00E0 */
  *out++ = 0x00;
  *out++ = 0xE0;
  varlen += 2;

  /** construct the TI-92 Graph Link header */

  memcpy (header.TI92, "**TI92**", 8);
  header.Unknown1[0] = 1;
  header.Unknown1[1] = 0;
  memcpy (header.Folder, "main\x0\x0\x0\x0", 8);

  getdate (&cdate);
  gettime (&ctime);
  sprintf (header.String, "Converted by TI92TC: %02d.%02d.%04d %02d:%02d   \x1",
           cdate.da_mon, cdate.da_day, cdate.da_year, ctime.ti_hour, ctime.ti_min);

  header.Unknown2[0] = 0x52;           /* TI-Graph-Link string signature ?? */
  header.Unknown2[1] = 0x00;
  header.Unknown2[2] = 0x00;
  header.Unknown2[3] = 0x00;

  /* get the name part of the filename (without fnsplit) */
  name = (tmp = strchr (OutFileName, ':'))  != NULL ? tmp+1 : OutFileName;
  name = (tmp = strrchr (OutFileName, '/')) != NULL ? tmp+1 : name;
  name = (tmp = strrchr (OutFileName, '\\')) >= name ? tmp+1 : name;
  strcpy (ename, name);
  if ((ext = strchr (ename, '.')) != NULL) /* remove extension, if exists */
    *ext = '\0';

  strncpy (header.Name, ename, 8);

  header.Unknown3[0] = 0x0B;
  header.Unknown3[1] = 0x00;
  header.Unknown3[2] = 0x00;
  header.Unknown3[3] = 0x00;

  header.GraphLinkLength = varlen + 0x58;
  header.Unknown4[0] = 0xA5;
  header.Unknown4[1] = 0x5A;
  header.Unknown4[2] = 0x00;
  header.Unknown4[3] = 0x00;
  header.Unknown4[4] = 0x00;
  header.Unknown4[5] = 0x00;

  checksum = GenerateCSum (outfile, varlen);

  /** write header+converted_text+checksum */
  if (fwrite (&header, 1, 0x56, OutHdl) == -1)
    WriteErr ();
  if (fwrite (outfile, 1, varlen, OutHdl) == -1)
    WriteErr ();
  fputc (checksum&0xFF, OutHdl);
  fputc (checksum>>8, OutHdl);

  puts ("done.");
  printf ("TI-92 Text Variable is %d bytes.\n", varlen);
}


/**
 * Checks, whether the input file is a TI-92 Graph Link file
 */
int IsGraphLinkFile ()
{
  return (memcmp (InFile, "**TI92**\x01", 10) == 0);
}

/**
 * Loads the input file
 */
void LoadInFile ()
{
  void MemError ()
  {
    fputs ("\nNot enough DPMI memory, to load input file.\n\n", stderr);
    exit (1);
  }
  void ReadError ()
  {
    fprintf (stderr, "\nCan't read input file : %s\n\n", strerror (errno));
    exit (1);
  }

  if ((InFile = malloc ((InFileLength = filelength (fileno (InHdl))) + 1)) == NULL)
    MemError ();

  if (fread (InFile, 1, InFileLength, InHdl) == -1)
    ReadError ();

  InFile[InFileLength] = '\0';
}

/**
 * Prints a short command line option help
 */
void Help ()
{
  char *helptext[] =
  {
    "",
    "TI92TC converts ASCII text files from and to TI-92 Graph Link Text Variables.",
    "",
    "Its syntax is:",
    "",
    "TI92TC <input-file> <output-file> [/c] [/w[NNN]]",
    "  <input-file>  - ASCII text or Graph Link Text Var; conversion mode will be",
    "                  set to 92T-->TXT or TXT-->92T, depending on its contents",
    "                  (auto detection)",
    "  <output-file> - Graph Link Var or ASCII Text (depends on conversion mode)",
    "  /c            - enable ASCII <---> TI-92 character set conversion",
    "  /w            - enable word wrapping; line length will be 37 characters in",
    "                  TXT-->92T mode, or 78 characters in 92T-->TXT mode",
    "  /wNNN         - enable word wrapping, and set line length to NNN characters",
    "",
    NULL
  };
  int I;

  for (I = 0; helptext[I]; I++)
    puts (helptext[I]);
}

/**
 * MAIN - processes command line options
 */
int main (int argn, char *argv[])
{
  int I;

  void InError ()
  {
    fprintf (stderr, "\nCan't open input file `%s' : %s\n\n", argv[1], strerror (errno));
    exit (1);
  }

  void OutError ()
  {
    fprintf (stderr, "\nCan't create output file `%s' : %s\n\n", argv[2], strerror (errno));
    exit (1);
  }

  void InvalidArg (int n)
  {
    fprintf (stderr, "\nInvalid argument: `%s'\n\n", argv[n]);
    Help ();
    exit (1);
  }

  puts ("\t----------------------------------------------------------------");
  puts ("\t       TI-92 Text Conversion v"VERSION_STR" - by David Khling 1998       ");
  puts ("\t----------------------------------------------------------------\n");

  if (argn < 3)
  {
    Help ();
    return (1);
  }

  /** open input and output file */
  if ((InHdl = fopen (argv[1], "rb")) == NULL)
    InError ();
  if ((OutHdl = fopen (OutFileName = argv[2], "wb+")) == NULL)
    OutError ();

  /** process the arguments */
  for (I = 3; I < argn; I++)
    if (argv[I][0] != '-' && argv[I][0] != '/')
      InvalidArg (I);
    else
      switch (argv[I][1])
      {
        case 'w':
          if (argv[I][2] == '\0')
            WordWrap = -1;
          else
            if ((WordWrap = atoi (&argv[I][2])) < 1 || WordWrap > 1000)
              InvalidArg (I);
          break;
        case 'c':
          if (argv[I][2] != '\0')
            InvalidArg (I);
          else
            CharConv = 1;
          break;
        default:
          InvalidArg (I);
      }

  LoadInFile ();

  /** convert... */
  if (IsGraphLinkFile ())
    Convert92T2TXT ();
  else
    ConvertTXT292T ();

  fclose (InHdl);
  fclose (OutHdl);
  free (InFile);
  return (0);
}
